#include <math.h>
#include "linmath.h"

//LAB03 for 6.302.1x, August 2016

//Likely User Modified Variables ******************************
//change if you want but shouldn't need to!

unsigned int deltaT = 1000;         // Sample period in microseconds.
int angleAverages = 5;
int past_size = 3;                 // interval for delta, larger=less noise, more delay.
unsigned long transferDt = 20000; // usecs between host updates  
float K1, K2, K3, directV, desiredV;
                                 
// ***************************************************************

//--------------------------------
//--------------------------------
//FILL IN YOUR MATRIX VALUES BELOW for Observer:
// The L vector for converging the observer. These gains converge quite slowly
vec3 L = {
    0.0,0.0, 0.0
};

// A matrix
mat3x3 A = {
  {0.0, 0.0, 0.0},
  {0,   0.0, 0.0},
  {0.0, 0.0, 0.0}
};
  
//B matrix (vector)
vec3 B = {
      0.0,
      0.0,
      0.0
          };

//C matrix (vector)
vec3 C = {
    0,   0,   0
};

//K vector (use gains you found that worked in lab 02.  
//Try to have some gain on K3 since it does help with overshoot
vec3 K = {
     0.0,   0.0,   0.0
};


float Ku = 0.0; //precompensator gain

float u = 0;
float y = 0; 

//estimated_state vector
vec3 xhat =  { 0, 0, 0 };

float rst = 0;
float real_est = 0;


// Teensy definitions
#define angleSensorPin  A9
#define pwmVoltagePin A0
#define motorVoltagePin  A1
#define motorOutPWM  4
#define monitorPin 2  
#define dacRes 12   // Teensy 12-bit dac resolution, max val 4095
#define dacMax 4095
#define adcRes 14
#define adcCenter 8192 //Teensy ADC set to 14 bits, 16384/2 = 8192
#define adcMax 16383

// Circuit Specific Definitions.
#define scaleVadc 5.0 // All signals 0->5 volts.
#define vDrive 5.0 // Driver peak voltage.
#define rShunt 0.5 // Resistor from motor to ground.
#define rMotor 0.5 // Motor series resistance.
#define Ke 0.5e-3; //back EMF constant

// Variables for Conversion
#define deg2rad 0.0175    
#define twopi 6.2831853  

//Rest of Setup:
bool first_time;
String config_message_30_bytes = "&A~DesiredAngV~5&C&S~DesiredAngV~A~-1~1~0.1&S~Direct~O~0~5~0.1&S~Reset~R~0~1~1&S~Real/Est~X~0~1~1&T~Theta~F4~-2.5~2.5&T~Omega_a~F4~-5~5&T~Omega_m~F4~0~5&T~X1~F4~-2~2&T~X2~F4~-2~2&T~X3~F4~-2~2&D~100&H~4&";
String config_message = config_message_30_bytes;

float rad2deg = 1.0/deg2rad;        // 180/pi
 
float dTsec = 1.0e-6*deltaT;       // The period in seconds. 
float scaleD = 1.0/(dTsec*past_size); // Divide deltas by interval.    
                             
float errorVintegral;                // Variable for integrating angle errors.
float integralMax = 200;            // Maximum value of the integral

int loopCounter;

// Storage for past values.
float pastAngleV[20];  // Should be larger array than past_size.
char buf[60];  // 

// Variables for loop control
uint32_t loop_counter;
int numSkip;  // Number of loops to skip between host updates.
elapsedMicros loopTime; // Create elapsed time variable to ensure regular loops.
unsigned int headroom;  // Headroom is time left before loop takes too long.
boolean switchFlag;

// Initializes past values.
void setup() {
  Serial.begin(115200);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
  
  // Set up inputs
  analogReadResolution(adcRes);
  pinMode(motorVoltagePin, INPUT);
  pinMode(pwmVoltagePin, INPUT);
  pinMode(angleSensorPin, INPUT);

  // Set up output
  analogWriteResolution(dacRes);
  pinMode(motorOutPWM, OUTPUT);
  analogWriteFrequency(motorOutPWM, 23437.5); // Teensy 3.0 pin 3 also changes to 23437.5 kHz
  analogWrite(motorOutPWM, LOW);
  
  // Frequency Monitor
  pinMode(monitorPin, OUTPUT);

  // Number of loops between transfers to host.
  numSkip = max(int((transferDt+100)/deltaT - 1),0); 

  first_time = false;

}

void loop() {  // Main code, runs repeatedly 
  // Reinitializes or updates from sliders on GUI.
  startup();
  
  // Make sure loop starts deltaT microsecs since last start
  unsigned int newHeadroom = max(int(deltaT) - int(loopTime), 0);
  headroom = min(headroom, newHeadroom);
  
  while (int(loopTime) < int(deltaT)) {};
  loopTime = 0;
  
  // Monitor Output should be a square wave with frequency = 1/(2*deltaT) 
  switchFlag = !switchFlag;
  digitalWrite(monitorPin, switchFlag);

  /*************** Section of Likely User modifications.**********************/
  // Read Angle, average to reduce noise.
  int angleI = 0;
  for(int i = 0; i < angleAverages; i++) angleI += analogRead(angleSensorPin);
  float angleV = scaleVadc*float(angleI- angleAverages*adcCenter)/float(angleAverages*adcMax);
  
  // Read motor voltage and pmw voltage
  float motorV = scaleVadc*float(analogRead(motorVoltagePin))/float(adcMax);
  float pwmV = scaleVadc*float(analogRead(pwmVoltagePin))/float(adcMax);
  float motor_current = (pwmV - motorV)/rShunt;
  float motor_bemf = motorV - rMotor*motor_current;  // Ignores L(di/dt).

  // Compute error, and error deltas from past.
  float errorV = (desiredV - angleV);
  float omegaArm = (angleV - pastAngleV[past_size-1])*scaleD;
// 1. UPDATE STATE
  vec3 temp1, temp2, temp3, temp4;
 
  // xhat[n+1] = A*xhat[n] + B*u[n] + L*(y[n] - C*xhat[n])::
  mat3x3_mul_vec3(temp1, A, xhat); // A*xhat
  vec3_scale(temp2,B,u); // B*u
  vec3_scale(temp3, L, (y-vec3_mul_inner(C,xhat))); // L*(y-yhat) = L*(y-C*xhat)
  //vec3_add(xhat,temp2,temp1); // A*xhat+B*u
  vec3_add(temp4,temp2,temp1); // A*xhat+B*u
  vec3_add(xhat, temp4,temp3); // xhatnew = A*xhat + B*u + L*(y-yhat) 


  if (real_est == 1){ //if using estimated (real_est==1):
    u = Ku*desiredV - K[0]*xhat[0] - K[1]*xhat[1] - K[2]*xhat[2] + directV;
  }else{ //if using actual states (real_est==0)
    u = Ku*desiredV - K[0]*angleV - K[1]*omegaArm - K[2]*motor_bemf + directV;
  }

  // Bound xhat.
  for (int i=0; i<3; i++){
    xhat[i] = min(max(xhat[i],-500),500);
  }
  if (rst == 1){
    for (int i=0; i<3; i++){
      xhat[i] = 0;   
    }
  }
  y = angleV;


  // calculate the motor command signal here:
   // Compute and write out motor command.
  float motorCmd = ((u)*dacMax/vDrive);
  int motorCmdInt = min(max(int(motorCmd), 0), dacMax);
  analogWrite(motorOutPWM,motorCmdInt);  

  // Update previous errors for next time.
  for (int i = past_size-1; i > 0; i--) pastAngleV[i] = pastAngleV[i-1];
  pastAngleV[0] = angleV;
  
  if (loopCounter == numSkip) {  // Lines below are for debugging.
    packStatus(buf, rad2deg*angleV, rad2deg*omegaArm, motor_bemf, rad2deg*xhat[0], rad2deg*xhat[1], xhat[2], float(headroom));
    Serial.write(buf,30);
    loopCounter = 0;
    headroom = deltaT;
  } else loopCounter += 1;

}

void init_loop() {
  // Initialize loop variables
  loopCounter = 0; 
  headroom = deltaT;
  // Zero past errors
  for (int i = past_size-1; i >= 0; i--) pastAngleV[i] = 0;
}


void processString(String inputString) {
char St = inputString.charAt(0);
  inputString.remove(0,1);
  float val = inputString.toFloat();
  switch (St) {
    case 'R':
      rst = val;
      break;
    case 'B': 
      K1 = val;
      break;
    case 'C':
      K2 = val;
      break;  
    case 'D':
      K3 = val;
      break;
    case 'U':
      Ku = val;
      break;
    case 'O':  
      directV = val;
      break;
    case 'A':
      desiredV = val*deg2rad;
      break;
    case 'X':
      real_est = val;
      break;
    case '~':
      first_time = true;
      break;
    default:
    break;  
  }
}

// Load the serial output buffer.
void packStatus(char *buf, float a, float b, float c, float d, float e, float f, float g) {
  
  // Start Byte.
  buf[0] = byte(0);
  int n = 1; 
  
  memcpy(&buf[n],&a,sizeof(a));
  n+=sizeof(a);
  memcpy(&buf[n],&b,sizeof(b));
  n+=sizeof(b);
  memcpy(&buf[n],&c,sizeof(c));
  n+=sizeof(c);
  memcpy(&buf[n],&d,sizeof(d));
  n+=sizeof(d);
  memcpy(&buf[n],&e,sizeof(e));
  n+=sizeof(e);
  memcpy(&buf[n],&f,sizeof(f));
  n+=sizeof(f);
  memcpy(&buf[n],&g,sizeof(g));
  n+=sizeof(g);

  // Stop byte (255 otherwise unused).
  buf[n] = byte(255); 
}

// Initializes the loop if this is the first time, or if reconnect sent
// from GUI.  Otherwise, just look for serial event.
void startup(){
  if (first_time) {
    while(Serial.available() > 0) Serial.read(); // Clear out rcvr.
    Serial.println(config_message);  // Send the configuration files.
    while (! Serial.available()) {}  // Wait for serial return.
    while(Serial.available() > 0) {   // Clear out rcvr.
        Serial.read();
        //char inChar = (char) Serial.read(); 
        //if (inChar == '\n') break;
    }
    init_loop();
    first_time = false;
  } else {
    serialEvent();
  }
}


// Simple serial event, only looks for disconnect character, resets loop if found.

void serialEvent() {
  String inputString = ""; 
  while (Serial.available()) {
    char inChar = (char)Serial.read();
    inputString += inChar;
    // if the incoming character is a newline, ready to process.
    if (inChar == '\n') {
      processString(inputString);
      inputString = "";
      break;
    }
  }
}
